#
# GroupedStats.py
#
import Stats

def echo2(str): sys.stderr.write(str + "\n")
def die(str, status=1): echo2(str); sys.exit(status)
def slice_by_list(seq, indices): map(lambda i: seq[i], indices)

class GroupedStats:
    """Compute statistics for specified variables (columns in a text line)
       while grouping the input lines over one or more key columns.
       Will sort using multiple keys only if requested (assumes prior sort)."""
    def __init__(self, keys, vars, config):
        self.reset(keys, vars, config)

    def reset(self, keys, vars, config):
        self.keys = keys      # list of pairs/triples (column_#, key_name, type='string')
                              # where type='string (default, lexicographic sort) or type='numeric'
        self.vars = vars      # list of pairs (column_#, var_name), vars must be numeric
        self.config = config  # configuration params in a dict

        self.n_levels = len(keys)    # number of grouping levels, max = 3
        if self.n_keys > 3: die("GroupedStats: Only 3 grouping levels (keys) allowed.")
        self.n_vars = len(vars)
        self.key_cols = map(lambda p: p[0], keys)
        self.key_names = map(lambda p: p[1], keys)
        self.key_types = map(lambda p: if len(p) > 2: p[2] \
                             else: 'string', keys)
        self.var_cols = map(lambda p: p[0], vars)
        self.var_names = map(lambda p: p[1], vars)
        self.prev_key_vals = None
        self.keys_label = join(self.key_names, '/') + ' = '
        self.out_lines = []
        for i in range(self.n_keys): stats[i] = []
        for i in range(self.n_keys):
            for j in range(self.n_vars):
                stats[i][j] = Stats()
        
    def readFrom(self, input_, sort=0):
        """Read input lines from input_: a filename string, a file object, or a sequence object."""
        if type(input_) == 'string':
            input = open(input_, "r")  # assume it's a filename
        else:
            input = input_
        if sort:
            if type(input) == 'file': input = input.readlines()
            input = sortLines(input)
        for line in input:
            self.addData(line)

    def sortLines(self, lines):
        aug_lines = []
        for line in lines:
            fields = split(line);
            tuple = slice_by_list(fields, self.key_cols);
            for i in range(self.n_keys):
                if self.key_types[i] == 'numeric': tuple[i] = float(tuple[i])
                else: tuple[i] = trim(tuple[i]);
            tuple.append(line)
            aug_lines.append(tuple)
         lines = map(lambda s: s[-1], aug_lines.sort())
         return lines

    def addData(self, line):
        fields = split(line)
        key_vals = slice_by_list(fields, self.key_cols)
        var_vals = slice_by_list(fields, self.var_cols)
        if self.prev_key_vals == None:
            self.prev_key_vals = key_vals
        else:
            for i in range(self.n_keys-1, -1, -1):
                if key_vals[i] != self.prev_key_vals[i]:
                # if key value has changed, output stats for that subgroup
                    for j in range(self.n_vars):
                        label = self.keys_label + join(prev_key_vals,'/') + ', ' +
                                  self.var_names[j] + ' stats'
                        line = str( stats[i][j].label(label) )
                        self.out_lines.append(line)
                    self.stats[i].reset()  # reset stats before adding current lines
                    self.prev_key_vals[i] = key_vals[i]
        for i in range(self.n_keys):
            for j in range(self.n_vars):
                self.stats[i][[j].add( float(self.var_cols[j]) )

    def __repr__(self):
        return join(self.out_lines, '\n') + '\n'


#if __name__ == '__main__':
    